#!python
# Copyright (c) 2022, Cisco Systems, Inc. and/or its affiliates.
# All rights reserved.
# See accompanying LICENSE file in apt2sbom distribution.
"""
Routines to read configuration.  Only necessary for web access (not
cli).

Current values supported with default values (if any):

   do_auth: True,  # otherwise don't auth
   passwd_file: "/etc/sbom.users",  # where to find the passwds
   include_pip: False, # pip results take a long time
   pregen_file: None # only open and read this file; otherwise gen.

"""

import json

def readconf(conffile):
    """
    read the configuration file, and return configuration elements in
    a dict.
    """

    try:
        with open(conffile,"r",encoding="utf-8") as conf_fp:
            conf=json.load(conf_fp)
    except OSError as j_error:
        print(f'Error: {j_error}')
        conf={}
    return conf



# Copyright (c) 2022, Cisco Systems, Inc. and/or its affiliates.
# All rights reserved.
# See accompanying LICENSE file in apt2sbom distribution.

"""
Routines to call from werkzeug to enable simple sbom web service.
"""

import json
from os import getuid
from flask import Flask,Response, request
from flask_httpauth import HTTPBasicAuth
from apt2sbom.dp2yaml import toyaml
from apt2sbom.dp2json import tojson
from apt2sbom.dp2cdx import tocyclonedx
from apt2sbom.readconf import readconf

conf=readconf("/etc/sbom.conf")

if getuid() == 0:
    raise RuntimeError("Not running as root")


if 'do_auth' in conf and conf['do_auth'] is True:
    if 'passwd_file' not in conf:
        raise ValueError('passwd_file')

    try:
        with open(conf['passwd_file'],"r",encoding="utf-8") as f:
            users = json.load(f)
    except OSError as pw_error:
        print("unable to load passwd file")
        exit(-1)

app = Flask(__name__)
app.url_map.strict_slashes = False
auth = HTTPBasicAuth()

if 'do_auth' in conf and conf['do_auth']:
    @auth.verify_password
    def verify_password(username, password):
        """ Basic password check """
        if username in users and \
           ( users.get(username) == password):
            return username
        return None

def get_sbom(pattern=None):
    """
    generate SBOM once we are authenticated.
    """
    if 'pre_gen' in conf and conf['pre_gen']:
        if 'sbom_type' in conf:
            sbom_mime=conf['sbom_type']
        else:
            raise ValueError("sbom_type required and not set.")
        try:
            with open(conf['pre_gen'],"r",encoding="utf-8") as sbom_fp:
                sbom=sbom_fp.read()
            return Response(sbom,mimetype=sbom_mime)
        except OSError as sbom_error:
            return Response(str(sbom_error),400)

    if ( "application/json" in request.accept_mimetypes or
         "application/spdx+json" in request.accept_mimetypes ):
        return Response(tojson(pattern),mimetype="application/spdx+json")
    if "application/vnd.cyclonedx+json" in request.accept_mimetypes:
        return Response(tocyclonedx(pattern),mimetype="application/vnd.cyclonedx+json")
    return Response(toyaml(pattern),mimetype="text/spdx")


if "do_auth" in conf and conf['do_auth']:
    @app.route('/',methods=['GET'])
    @auth.login_required
    def return_sbom():
        """ Return an SBOM with no params"""
        return get_sbom(None)

    @app.route('/<pattern>',methods=['GET'])
    @auth.login_required
    def search_sbom(pattern = None):
        """ return sbom with a search param """
        pattern = '.*' + pattern + '.*'
        return get_sbom(pattern)

else:
    @app.route('/',methods=['GET'])
    def return_sbom():
        """ Return an SBOM with no params"""
        return get_sbom(None)

    @app.route('/<pattern>',methods=['GET'])
    def search_sbom(pattern = None):
        """ return sbom with a search param """
        pattern = '.*' + pattern + '.*'
        return get_sbom(pattern)



#!python
# Copyright (c) 2022, Cisco Systems, Inc. and/or its affiliates.
# All rights reserved.
# See accompanying LICENSE file in apt2sbom distribution.
"""
routine to convert apt and pip information to CycloneDX.
"""
import re
import json
from uuid import uuid4
from datetime import datetime
from apt.cache import Cache
from .piplist import getpip

def tocyclonedx(pattern = None,dopip=False):
    """
    Routine to convert apt information to CycloneDx.
    """
    pkgs = [ ]
    deps = []

    sbom = {
        "bomFormat": "CycloneDX",
        "specVersion": "1.4",
        "serialNumber": "urn:uuid:" + str(uuid4()),
        "version": 1
        }
    meta = {
        "timestamp": re.sub(r'[.].*$','',str(datetime.now().isoformat())) + 'Z',
        "tools" : [ {
            "vendor" : "Eliot Lear",
            "name" : "apt2sbom",
            } ],
        "component" : {
            "type" : "device",
            "name" : "ubuntu-derived-system",
            "version" : "1"
            },
        "licenses" : [ { "license" : {
            "id" : "BSD-3-Clause"
            }
        } ]
      }
    sbom['metadata']= meta

    cache=Cache()

    for pkg in cache:
        if not pkg.is_installed:
            continue
        if pattern:
            if not re.match(pattern,pkg.name):
                continue
        ver = pkg.installed
        pack = { }
        pack['type'] = 'application'
        pack["name"]=pkg.name
        rec_info=ver.record
        pack["bom-ref"] = pkg.name
        pack["version"] = ver.version
        pack["purl"] = "pkg:deb/ubuntu/" + pkg.name + "@" + ver.version +\
            "?arch=" + ver.architecture
        pack["supplier"] = { "name" : rec_info['Maintainer'] }
        if not ver.uri is None and not ver.uri == "":
            pack['supplier']['url']= [ ver.uri ]

        hashes= []
        try:
            hashes.append({ "alg" : 'SHA-256',
                       'content' : ver.sha256 })
        except SystemError:
            pass
        try:
            hashes.append({ "alg" : 'SHA-1',
                       'content' : ver.sha1 })
        except SystemError:
            pass

        try:
            hashes.append({ "alg" : 'MD5',
                       'content' : ver.md5 })
        except SystemError:
            pass

        if hashes:
            pack['hashes'] = hashes

        if ver.homepage != '':
            pack['externalReferences'] = [ {
                "url" : ver.homepage,
                "type" : "website"
            } ]

        if ver.dependencies:
            dep = { "ref" : pack["bom-ref"] }
            dees = []

            for dep_ent in ver.dependencies:
                tname=re.sub(':any$','',dep_ent[0].name)
                if tname in cache and cache[tname].is_installed and\
                   not tname in dees:
                    dees.append(tname)

            if dees:
                dep["dependsOn"] = dees
                deps.append(dep)

        pkgs.append(pack)

    if dopip:
        pips = getpip()
        for pip in pips:
            if pattern:
                if not re.match(pattern,pip['Name']):
                    continue
            pack={}
            pack["name"]=pip["Name"] + ".pip"
            pack['type']="application"
            pack["version"] = pip["Version"]
            pack["purl"] = "pkg:pypi/" + re.sub('_','-',pack["name"].lower()) +\
                "@" + pack["version"]
            pack["bom-ref"] = pack["name"]
            pack["supplier"] = {
                "name" : pip["Author"]
                }
            try:
                pack["externalReferences"] = [ {
                    "url" : pip["homepage"],
                    "type" : "website"
                    }]
            except KeyError:
                pass
            pkgs.append(pack)

    sbom['components'] = pkgs
    sbom['dependencies'] = deps
    return json.dumps(sbom)



#!python3
# Copyright (c) 2022, Cisco Systems, Inc. and/or its affiliates.
# All rights reserved.
# See accompanying LICENSE file in apt2sbom distribution.
"""
Basic initialization for apt2sbom.
"""


from .dp2json import *
from .dp2yaml import *
from .dp2cdx import *
from .apt2sbom import *
from .piplist import *



#!python
# Copyright (c) 2022, Cisco Systems, Inc. and/or its affiliates.
# All rights reserved.
# See accompanying LICENSE file in apt2sbom distribution.
"""
Routines to generate lists of python modules installed.
"""

import re
import pip_api
from pip_api._call import call

# return an array of global packages.


def getpip():
    """
    Generate a pip package list.
    """
    pkglist = []
    try:
        pip=pip_api.installed_distributions()
    except SystemError:
        return []

    for dep in pip:
        args=["show",pip[dep].name]
        attrs=re.split('\n',call(*args))

        if not attrs == []:
            entry={}
            for attr in attrs:
                tlv=re.split(': ',attr)
                if len(tlv) == 2:
                    entry[tlv[0]] = tlv[1]
            pkglist.append(entry)
    return pkglist



#!python
# Copyright (c) 2022, Cisco Systems, Inc. and/or its affiliates.
# All rights reserved.
# See accompanying LICENSE file in apt2sbom distribution.
"""
Convert apt and pip information to SPDX JSON format.
"""

# import json
from socket import gethostname
from datetime import datetime
import re
import json
from apt.cache import Cache
from .piplist import getpip

def tojson(pattern = None,dopip=False):
    """
    Convert APT information to SPDX JSON.
    """

    sbom = { }
    cinfo = { }
    pkgs = [ ]
    pkgids = [ ]
    deps = {}
    rels = []
    sbom = {
        "spdxVersion" : "SPDX-2.2",
        "SPDXID" : "SPDXRef-DOCUMENT",
        "dataLicense" : "CC0-1.0",
        "name" : "apt2sbom-" + gethostname(),
        "documentNamespace" : "https://" + gethostname() + "/.well-known/transparency/sbom"
    }
    cinfo["creators"] =  [ "Tool: apt2sbom-ubuntu-1.0" ]
    cinfo["created"] = str(re.sub(r'..*$','',datetime.now().isoformat())) + 'Z'

    sbom['creationInfo']= cinfo


    cache=Cache()

    for pkg in cache:
        if not pkg.is_installed:
            continue
        if pattern:
            if not re.match(pattern,pkg.name):
                continue
        ver = pkg.installed
        rec_info=ver.record
        pack = { }
        pack["name"]=pkg.name
        pack["SPDXID"] = "SPDXRef-apt2sbom." + pkg.name
        pkgids.append(pack["SPDXID"])
        pack["versionInfo"] = re.sub("[:~]","-",ver.version)
        pack["filesAnalyzed"] = False
        pack["supplier"] = "Organization: " + rec_info['Maintainer']
        pack["homepage"]= ver.homepage
        hashes= []
        try:
            hashes.append({ "algorithm" : 'SHA256',
                       'checksumValue' : ver.sha256 })
        except SystemError:
            pass

        try:
            hashes.append({ "algorithm" : 'SHA1',
                       'checksumValue' : ver.sha1 })
        except SystemError:
            pass

        try:
            hashes.append({ "algorithm" : 'MD5',
                       'checksumValue' : ver.md5 })
        except SystemError:
            pass

        if hashes:
            pack['checksums'] = hashes

        if ver.uri:
            pack["downloadLocation"]= ver.uri
        else:
            pack["downloadLocation"]=  "http://spdx.org/rdf/terms#noassertion"
        if ver.dependencies:
            deps[pkg.name] = []
            for dep in ver.dependencies:
                tname=re.sub(':any$','',dep[0].name)
                if tname in cache and cache[tname].is_installed:
                    deps[pkg.name].append(tname)
            if deps[pkg.name] == []:
                deps.pop(pkg.name)
        pack["licenseConcluded"] = "NOASSERTION"
        pack["licenseDeclared"] = "NOASSERTION"
        pack["copyrightText"] = "NOASSERTION"
        pkgs.append(pack)
    if dopip:
        pips = getpip()
        for pip in pips:
            if pattern:
                if not re.match(pattern,pip['Name']):
                    continue
            pack={}
            pack["name"]=pip['Name']
            pack["versionInfo"] = pip['Version']
            pack["SPDXID"] = "SPDXRef-apt2sbom.pip." + pip['Name']
            pack["supplier"] = "Organization: " + pip['Author'] + ' <'\
                + pip['Author-email'] + '>'
            try:
                pack["homepage"] = pip['home-page']
            except KeyError:
                pass
            pack["filesAnalyzed"] = False
            pack["downloadLocation"] = "http://spdx.org/rdf/terms#noassertion"
            pack["licenseConcluded"] = "NOASSERTION"
            pack["licenseDeclared"] = "NOASSERTION"
            pack["copyrightText"] = "NOASSERTION"
            pkgs.append(pack)
            pkgids.append(pack["SPDXID"])

    sbom['packages'] = pkgs
    sbom['documentDescribes'] = pkgids
    for pname in deps:
        for dep in deps[pname]:
            rec_info = {
                  'spdxElementId' : "SPDXRef-apt2sbom." + pname,
                  'relationshipType' : 'DEPENDS_ON',
                  'relatedSpdxElement' : "SPDXRef-apt2sbom." + dep
            }
            rels.append(rec_info)

    sbom['relationships']= rels
    return json.dumps(sbom)



#!python
# Copyright (c) 2022, Cisco Systems, Inc. and/or its affiliates.
# All rights reserved.
# See accompanying LICENSE file in apt2sbom distribution.
"""
Provide CLI functionality for apt2sbom library.
"""

import argparse
from apt2sbom import tojson,toyaml,tocyclonedx

def cli():
    """
    Function to call CLI routines to invoke apt2sbom.
    """

    parser= argparse.ArgumentParser(description="generate SPDX file from APT inventory")
    group=parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-j','--json',help="Generate JSON SPDX output", action='store_true')
    group.add_argument('-y','--yaml',help="Generate YAML SPDX output", action='store_true')
    group.add_argument('-c','--cyclonedx',\
                       help="Generate CycloneDX JSON output",\
                       action='store_true')
    parser.add_argument('-p','--pip',help="Include PIP files",action='store_true')
    args=parser.parse_args()

    if args.json:
        out=tojson(dopip=args.pip)
    elif args.yaml:
        out=toyaml()
    else:
        out=tocyclonedx(dopip=args.pip)
    print(out)



# Copyright (c) 2022, Cisco Systems, Inc. and/or its affiliates.
# All rights reserved.
# See accompanying LICENSE file in apt2sbom distribution.
"""
Routines to convert APT information -> SPDX YAML.
"""

import re
from datetime import datetime
from socket import gethostname
from apt.cache import Cache

def toyaml(pattern = None):
    """
    Function to convert APT information to YAML.
    """
    res="SPDXVersion: SPDX-2.2\nDataLicense: CC0-1.0\nSPDXID: SPDXRef-DOCUMENT"
    res+="DocumentName: apt2sbom-" + gethostname() + "\n"
    res+="DocumentNamespace: https://" + gethostname() + "/.well-known/transparency/sbom\n"
    res+="Creator: Tool: apt2sbom-ubuntu-1.0\n"
    res+="Created: " + str(re.sub(r'..*$','',datetime.now().isoformat())) + 'Z\n\n'

    cache=Cache()
    deps = {}
    for pkg in cache:
        if not pkg.is_installed:
            continue
        if pattern:
            if not re.match(pattern,pkg.name):
                continue
        ver = pkg.installed
        rec_info=ver.record
        clean_ver = ver.version.replace("[:~]","-")
        res+="SPDXID: SPDXRef-apt2sbom." + pkg.name + "." + \
            clean_ver +"\n"
        res+="PackageName: " + pkg.name + "\n"
        res+="PackageVersion: " + clean_ver + "\n"
        res+="FilesAnalyzed: false" + "\n"
        res+="PackageSupplier: Organization: " +  rec_info['Maintainer'] + "\n"
        if not ver.homepage == '':
            res+="PackageHomePage: "+ ver.homepage + "\n"
        try:
            res+="PackageChecksum: SHA256: " + ver.sha256 + "\n"
        except SystemError:
            pass
        if ver.uri and not ver.uri == "":
            res+="PackageDownloadLocation: " + str(ver.uri) + "\n"
        else:
            res+="PackageDownloadLocation: NOASSERTION\n"
        res+="PackageLicenseConcluded: NOASSERTION" + "\n"
        res+="PackageLicenseDeclared: NOASSERTION" + "\n"
        res+="PackageCopyrightText: NOASSERTION" + "\n\n"
        if not ver.dependencies == []:
            deps[pkg.name] = []
            for dep in ver.dependencies:
                tname=re.sub(':any$','',dep[0].name)
                if tname in cache and cache[tname].is_installed:
                    deps[pkg.name].append(tname)
            if deps[pkg.name] == []:
                deps.pop(pkg.name)
    for k in deps:
        for dep in deps[k]:
            res += "Relationship: " + "SPDXRef-dpkg2spdx." + k + \
               " DEPENDS_ON " + "SPDXRef-dpkg2spdx." + dep + "\n"
    res+="\n"
    return res
